/*
Copyright (c) 2013-2014 Gustav Hartvigsson

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

#include "Map.h"
#include "LinkedList.h"
#include <stdlib.h>

struct
SMap {
  size_t len; // Length
  SDynamicArray * array; /*  This is what holds the buckets. These buckets are
                          * in turn also SDynamicArrays; or in template speak:
                          * SDynamicArray<SLinkedList<SMapItem*>*>*
                          */

  CompFunc is_equal; /*  method to check if items are equal. */
  HashFunc key_hash_func;
  FreeFunc free_key;
  FreeFunc free_value;
};

void
_s_map_internal_free_map_items_for_each (SMap * self,
                                         SMapItem * item,
                                         sboolean free_data);

SLinkedList *
_s_map_internal_find_bucket (SMap * self,
                             spointer key);

SMapItem *
_s_map_internal_find_item_in_bucket (SMap * self,
                                     SLinkedList * bucket,
                                     spointer key);

void
_internal_s_map_free_buckets (SLinkedList * bucket);

SMapItem *
s_map_item_new (void * key, void * value) {
  SMapItem * self = s_malloc (sizeof (SMapItem));
  self->key = key;
  self->value = value;

  return self;
}

void
s_map_item_free (SMapItem * self,
                 FreeFunc free_key,
                 FreeFunc free_value) {
  s_dbg_print ("Freeing Item!");
  if (free_key) {
    s_dbg_print ("Freeing Key");
    free_key (self->key);
  }
  if (free_value) {
    s_dbg_print ("Freeing Value");
    free_value (self->value);
  }
  s_free (self);
}


SMap *
s_map_new (CompFunc comp_func,
           HashFunc key_hash_func,
           FreeFunc free_key,
           FreeFunc free_value) {
  s_dbg_print ("Freeing SMap");
  SMap * self = s_malloc (sizeof (SMap));
  self->len = 0;

  self->is_equal = comp_func;

  /* free_* functions need to be checked if they are null and set the pointer
   * to free or s_base_object_free ()... Have to decide which...?
   */
  self->free_key = free_key;
  self->free_value = free_value;

  /*
   * Creating a place to store the items.
   */
  self->array = s_dynamic_array_new (S_MAP_DEFAULT_NUMBER_OF_BUCKETS,
                                     FREEFUNC(_internal_s_map_free_buckets));

  /* if no func is set we have to use some other metod of
   */
  if (key_hash_func == NULL){
    self->key_hash_func = s_hash_object;
  } else {
    self->key_hash_func = key_hash_func;
  }

  return self;
}

void
s_map_free (SMap * self, sboolean free_data) {

  s_map_for_each (self,
                  FOREACHFUNC(_s_map_internal_free_map_items_for_each),
                  &free_data);

  s_dynamic_array_free (self->array, TRUE);
  s_free (self);
}

void
s_map_add (SMap * self, spointer key, spointer value) {
  SDynamicArray * array = self->array;
  SMapItem * item = s_map_item_new (key, value);

  /* We have to generate a key to use as an index in the DynamicArray. */
  HashFunc hash_func = self->key_hash_func;
  hash_t hash = hash_func (key);
  /* We need to mod it with the max size of the array. */
  hash = hash % S_MAP_DEFAULT_NUMBER_OF_BUCKETS;

  if (self->len == 0) {
    /* Since we know that there are no items in the array we can skip the
     * check if the new place is taken. and just and an array to it.
     */
    s_dbg_print ("Adding bucket to bucket array");
    SLinkedList * bucket = s_linked_list_new (NULL);
    s_dbg_print ("1) Appending item to linked list");
    s_dynamic_array_set (array, hash, bucket);
    s_linked_list_append (bucket, item);
  } else {
    s_dbg_print ("SMap size in more than zero.");
    /* Figure out if the bucket exists. */
    SLinkedList * bucket = _s_map_internal_find_bucket (self, key);

    if (bucket == NULL) {
      s_dbg_print ("Bucket does not exist");
      bucket = s_linked_list_new (NULL);
      s_dynamic_array_set (array, hash, bucket);
    }
    
    SMapItem * tmp_item = _s_map_internal_find_item_in_bucket (self, bucket,
                                                              key);
    
    if (tmp_item) {
      s_dbg_print ("Key already exists in SMap... replacing it.\n");
      /* Since we know that the buckets current item points to the item (it
       * returns when it has found the item.) We can just remove it.
       */
      s_linked_list_remove_current (bucket, TRUE);
    }
    s_dbg_print ("2) Appending item to linked list");
    s_linked_list_append (bucket, item);
  }
  self->len++;
}

spointer
s_map_get (SMap * self, spointer key) {
  return s_map_get_item(self, key)->value;
}

SMapItem *
s_map_get_item (SMap * self, spointer key) {
  SLinkedList * bucket = _s_map_internal_find_bucket (self, key);
  if (bucket) {
    SMapItem * item = _s_map_internal_find_item_in_bucket (self, bucket, key);
    if (item) {
      return item;
    }
  }
  return NULL;
}

void
s_map_remove (SMap * self, spointer key) {
  SLinkedList * bucket = _s_map_internal_find_bucket (self, key);
  if (bucket) {
    s_linked_list_head (bucket);
    SMapItem * item = _s_map_internal_find_item_in_bucket (self, bucket, key);
    if (item) {
      s_map_item_free (item, self->free_key, self->free_value);
      s_linked_list_remove_current (bucket, FALSE);
      self->len--;
      return;
    }
  }
  s_dbg_print ("Could not find item in SMap.\n");
}

void
s_map_for_each (SMap * self, ForEachFunc foreach_func, spointer user_data) {
  s_dbg_print ("Running For Each on SMap");
  for (size_t i = 0; i <= s_dynamic_array_last_item (self->array); i++) {
    SLinkedList * bucket = s_dynamic_array_get (self->array, i);
    if (bucket) {
      s_dbg_print ("Found bucket");
      s_linked_list_head (bucket);
      s_dbg_print ("No of items in list: %zd.", s_linked_list_len (bucket));
      do {
        SMapItem * item = s_linked_list_get_current (bucket);
        foreach_func (self, item, user_data);
      } while (s_linked_list_next (bucket));
    }
  }
}

SLinkedList *
_s_map_internal_find_bucket (SMap * self,
                             spointer key) {
  s_dbg_print ("Looking for bucket");
  SLinkedList * ret_val = NULL;

  HashFunc hash_func = self->key_hash_func;
  hash_t hash = hash_func (key)  % S_MAP_DEFAULT_NUMBER_OF_BUCKETS;

  ret_val = s_dynamic_array_get (self->array, hash);

  return ret_val;
}


SMapItem *
_s_map_internal_find_item_in_bucket (SMap * self,
                                     SLinkedList * bucket,
                                     spointer key) {
  s_dbg_print ("Looking for item in bucket");
  s_linked_list_head (bucket);
  do {
    SMapItem * item = SMAPITEM (s_linked_list_get_current (bucket));

    if (item && self->is_equal (item->key, key)) {
      s_dbg_print ("Found Item in Bucket.");
      return item;
    }
  } while (s_linked_list_next (bucket));

  return NULL;
}

void
_s_map_internal_free_map_items_for_each (SMap * map,
                          SMapItem * item,
                          sboolean free_data) {
  s_dbg_print ("runnnig: _s_map_internal_free_map_items_for_each");
  if (free_data) {
    s_dbg_print ("Freeing K:V pair");
    s_map_item_free (item, map->free_key, map->free_value);
  } else {
    s_free (item);
  }
}

void
_internal_s_map_free_buckets (SLinkedList * bucket) {
  s_linked_list_free (bucket, FALSE);
}

